#ifndef __CIRC_BUFFER_
#define __CIRC_BUFFER_
#include <math.h>
#include <assert.h>
#include  <stdlib.h>
#include "my_defs.h"


template <class Element>
class CircularBuffer
{
public:

  // Constructor
  CircularBuffer(int maxCapacity, int delay, bool initBuffer, Element initValue)
  {
    m_Buffer = new Element[maxCapacity];
    
    // Sanity checks
    assert((void*)m_Buffer != NULL);
    assert(delay < maxCapacity);
    assert(maxCapacity > 0);
    assert(delay >=0);

    if (initBuffer) {
      // Initialize buffer
      for (int i=0; i<maxCapacity; i++) {
        m_Buffer[i] = initValue;
      }
    }

    m_count = 0;
    m_maxCapacity = maxCapacity;
    m_delay = delay;

    // Set up write index
    m_writeIndex = 0;

    // Set up read index to point 'delay' slots behind the write index
    m_readIndex = m_writeIndex - delay;
    if (m_readIndex < 0 ) {
      m_readIndex += maxCapacity;
    }
    DEBUG_PRINTF("CircularBuffer constructor: writeIndex = %d, m_readIndex = %d\n",m_writeIndex, m_readIndex);
  }
 
  //Destructor
  ~CircularBuffer(void)
  {
    delete[] m_Buffer;
  }

  // Inserts the given element into the start of the buffer, and reads
  // out an element from the end of the buffer.
  // This keeps the read pointer exactly 'm_delay' slots behind the
  // write pointer.
  Element pushPop(Element pushedElement)
  {
    // Note: must do the write before the read in order
    // to handle a delay of zero.

    /************** Write **********************/

    m_Buffer[m_writeIndex] = pushedElement;

    // Increment write index
    m_writeIndex++;
    // Wrap the index at the end of the array
    if (m_writeIndex >= m_maxCapacity) {
      m_writeIndex = 0;
    }

    /************** Read **********************/
    Element poppedElement = m_Buffer[m_readIndex];
    // Increment read index
    m_readIndex++;
    // Wrap the index at the end of the array
    if (m_readIndex >= m_maxCapacity) {
      m_readIndex = 0;
    }


    return poppedElement;
  }


  // Returns a copy of the Element written into the buffer 'numSlots' writes ago
  // (where n is specified using the slots argument)
  // Does not modify the buffer in any way
  // The slots parameter must be smaller than maxCapacity
  Element peekBack(int numSlots)
  {
    int index;
    assert(numSlots < m_maxCapacity);

    index = m_writeIndex-1-numSlots;
    if (index < 0) {
      index+= m_maxCapacity;
    }
    return m_Buffer[index];
  }


  bool setDelay(int delay)
  {
    if(delay >= m_maxCapacity) {
      return false;
    }

    // Set up read index to point 'delay' slots behind the write index
    m_readIndex = m_writeIndex - delay;
    if(m_readIndex < 0 ) {
      m_readIndex += m_maxCapacity;
    }

    // save off the new delay for future reference
    m_delay = delay;

    DEBUG_PRINTF("CircularBuffer setDelay: new delay = %d"
      "m_writeIndex = %d, m_readIndex = %d\n", m_delay, m_writeIndex, m_readIndex);
    
    return true;
  }


  void dump(void)
  {
    printf("Dumping content of buffer:\n");
    if(m_Buffer) {
      for(int i=0; i<m_maxCapacity; i++) {
        printf("Buffer[%d] = %f\n",i,(double)m_Buffer[i]);
      }
    }

    printf("m_dealy = %d\n",m_delay);
    printf("m_maxCapacity = %d\n",m_maxCapacity);
    //printf("m_count = %d\n",m_count);
    printf("m_writeIndex = %d\n",m_writeIndex);
    printf("m_readIndex = %d\n",m_readIndex);
  }


private:
  Element * m_Buffer;
  int m_delay;
  int m_maxCapacity;
  int m_count;
  int m_writeIndex;
  int m_readIndex;  
};




#endif //__CIRC_BUFFER_

