//
//  ES1Renderer.m
//  SonicSlingShot
//
//  Created by Roy Fejgin on 1/29/10.
//  Copyright __MyCompanyName__ 2010. All rights reserved.
//

#import "ES1Renderer.h"

// Mopho API
#include "mo_gfx.h"
#include "mo_audio.h"
#include "mo_touch.h"
#include "mo_audio.h"
// STK
#include "Stk.h"

// STL
#include <list>

// project includes
#include "my_defs.h"
#include "my_util.h"

using namespace std;

/*********************************************************************************
  Macros
 **********************************************************************************/

// Macros
#define SAMPLE_RATE    24000
#define FRAME_SIZE     128
#define NUM_CHANNELS   2


/*********************************************************************************
 Globals
 **********************************************************************************/


// Sine wave parameters
float g_carrierFreq = 440.0;
float g_carrierOffset = 0.0;
float g_targetCarrierOffset = 0.0;

// Texture name (handle)
GLuint g_texture[3];

ES1Renderer* g_view = NULL;


/*********************************************************************************
  Function Implementations
 **********************************************************************************/

// Touch callback
void touchCallback( NSSet * touches, UIView * view, 
                   const std::map<int, MoTouchTrack> * touchPts,
                   void * data)
{
    // Forward the call to the main view
    [g_view touchEvent: touchPts : view];
}


// Audio callback
void audioCallback( Float32 * buffer, UInt32 numFrames, void * userData )
{    
    for (int i=0; i<numFrames; i++) {
        // Get sample from the world simulation
        GLfloat output = [g_view renderAudio];

        // Output to all channels
        for (int j=0; j < NUM_CHANNELS; j++) {
            buffer[2*i+j] = output;
        }
    }
}

/*********************************************************************************
 ES1Renderer implementation
 **********************************************************************************/

@implementation ES1Renderer

// Create an ES 1.1 context
- (id) init
{
	if (self = [super init])
	{
		context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];
        
        if (!context || ![EAGLContext setCurrentContext:context])
		{
            [self release];
            return nil;
        }
		
		// Create default framebuffer object. The backing will be allocated for the current layer in -resizeFromLayer
		glGenFramebuffersOES(1, &defaultFramebuffer);
		glGenRenderbuffersOES(1, &colorRenderbuffer);
		glBindFramebufferOES(GL_FRAMEBUFFER_OES, defaultFramebuffer);
		glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
		glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_RENDERBUFFER_OES, colorRenderbuffer);
	}
    
	stk::Stk::setSampleRate( SAMPLE_RATE );

    // Create main objects
    pModel = new WorldModel(SAMPLE_RATE);
    pUISling = new UI_SlingShot(pModel, g_texture[0]);
    pUIProj = new UI_Projectiles(pModel, g_texture[1], g_texture[2]);
    
    // Load textures
    [self loadTextures];
    
    g_view = self;
    
    // init audio
    bool result = MoAudio::init(SAMPLE_RATE, FRAME_SIZE, NUM_CHANNELS);
    if (!result) {
        NSLog(@"Failed to init MoAudio!");
        return nil;
    }
    
    // start audio
    result = MoAudio::start(audioCallback, NULL);
    if (!result) {
        NSLog(@"Failed to start MoAudio!");
        return nil;
    }
    
    // Register touch callback
    MoTouch::addCallback(touchCallback, self);
    
	return self;
}

- (void) render
{
    // Replace the implementation of this method to do your own custom drawing

    
	// This application only creates a single context which is already set current at this point.
	// This call is redundant, but needed if dealing with multiple contexts.
    [EAGLContext setCurrentContext:context];
    
	// This application only creates a single default framebuffer which is already bound at this point.
	// This call is redundant, but needed if dealing with multiple framebuffers.
    glBindFramebufferOES(GL_FRAMEBUFFER_OES, defaultFramebuffer);
    glViewport(0, 0, backingWidth, backingHeight);
    
    // Set orthographic projection
    // Note: I've modified the ortho() function to compensate for 
    // assymetric aspect ratio of the iPhone's screen. 
    MoGfx::ortho();
    
    glMatrixMode(GL_MODELVIEW);    
	glLoadIdentity();
	
    glClearColor(0.5f, 0.5f, 0.5f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);

    // Advance simulation
    pModel->advance();
    
    // Draw slingshot
    pUISling->draw();
    
    // Draw projectiles and explosions
    pUIProj->draw();
    

	// This application only creates a single color renderbuffer which is already bound at this point.
	// This call is redundant, but needed if dealing with multiple renderbuffers.
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
    [context presentRenderbuffer:GL_RENDERBUFFER_OES];
}
- (GLfloat) renderAudio
{
    return pModel->tick();
}

- (void) touchEvent:(const std::map<int, MoTouchTrack>*) touchPts: (UIView*) view
{
    pModel->touchEvent(*touchPts, view);
}

- (BOOL) resizeFromLayer:(CAEAGLLayer *)layer
{	
	// Allocate color buffer backing based on the current layer size
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
    [context renderbufferStorage:GL_RENDERBUFFER_OES fromDrawable:layer];
	glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_WIDTH_OES, &backingWidth);
    glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_HEIGHT_OES, &backingHeight);
	
    if (glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES) != GL_FRAMEBUFFER_COMPLETE_OES)
	{
		NSLog(@"Failed to make complete framebuffer object %x", glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES));
        return NO;
    }
    
    return YES;
}

- (void) dealloc
{
	// Tear down GL
	if (defaultFramebuffer)
	{
		glDeleteFramebuffersOES(1, &defaultFramebuffer);
		defaultFramebuffer = 0;
	}

	if (colorRenderbuffer)
	{
		glDeleteRenderbuffersOES(1, &colorRenderbuffer);
		colorRenderbuffer = 0;
	}
	
	// Tear down context
	if ([EAGLContext currentContext] == context)
        [EAGLContext setCurrentContext:nil];
	
	[context release];
	context = nil;
	
	[super dealloc];
}

// Load textures for the entire program
- (void) loadTextures
{
    // enable texture mapping
    glEnable( GL_TEXTURE_2D );
    // enable blending
    glEnable( GL_BLEND );
    // blend function
    // glBlendFunc( GL_ONE, GL_ZERO );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    
    // generate texture name
    glGenTextures( 4, g_texture );

    // bind the texture
    glBindTexture( GL_TEXTURE_2D, g_texture[0] );
    // setting parameters
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    // Touch-point texture
    MoGfx::loadTexture( @"circle5", @"png" );
    
    // bind the texture
    glBindTexture( GL_TEXTURE_2D, g_texture[1] );
    // setting parameters
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    // Projectile texture
    MoGfx::loadTexture( @"circle8", @"png" );
    
    // bind the texture
    glBindTexture( GL_TEXTURE_2D, g_texture[2] );
    // setting parameters
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    // Explosion texture
    MoGfx::loadTexture( @"explosion5", @"png" );
    glBindTexture( GL_TEXTURE_2D, g_texture[0] );

    
}

@end
