/*
 *  ProjectileModel.h
 *  Model of a set of projectiles, i.e. a set of objects bouncing around
 *  a 'world' whose boundaries are the iPhones screen.
 *  The projectiles bounce off each other and off walls following
 *  (my very basic understaning of) physical laws.
 *
 *  If the projectiles collide with a total velocity exceeding a threshold,
 *  they cause an explosion, and both are eliminated from the world.
 *
 *  Also, after a predetermined number of collisions ('TTL'), the projectiles
 *  are allowed to drift out of the screen boundaries, and are then destroyed. 
 *
 *  Created by Roy Fejgin on 2/3/10.
 *  Copyright 2010 __MyCompanyName__. All rights reserved.
 *
 */


#ifndef __PROJ_MODEL_H__
#define __PROJ_MODEL_H__


#import "my_defs.h"
#import "mo_gfx.h"
#import <list>
using namespace std;

// Projectile
typedef struct Projectile {
    Vector3D location;
    Vector3D direction;
    Float32 radius;
    int TTL;
} Projectile;

// Collision type
typedef enum ColType
{
    COL_WALL,
    COL_EXPLOSION,
    COL_BOUNCE,
    COL_LAST
} ColType;

typedef struct Collision {
    CGPoint location;
    ColType type;    
    // anything else?
} Collision;

// Specific projectile data exported to the users of this model.
typedef struct ExportedProjectile
{
    CGPoint location;
    Float32 radius; 
} ExportedProjectile;

// Collision metadata
typedef struct ColMetaData
{
    bool bounceCollision;
    bool expCollision;
} ColMetaData;

class ProjectileModel {
    
public:
    // Constructor
    ProjectileModel(GLfloat xMax, GLfloat yMax) : m_xMax(xMax), m_yMax(yMax) {}
    
    void addProjectile(CGPoint sling1, CGPoint sling2, CGPoint proj);
    
    // Get list of all projectiles in the model
    list<ExportedProjectile> getProjectiles();
    
    // Get list of collisions generated during the last call to advance().
    list<Collision> getCollisions();
    
    // Advances the simulation by one time unit.
    ColMetaData advance();
    
private:
    bool simulateCollision(Projectile* proj1, Projectile* proj2, Collision* where);
    
private:
    // number of collisions until elimination
    static const int DefaultTTL = 10;
    // velocity threshold. Exceeding it in a collision to triggers an explosion.
    static const Float32 ExplosionThresh = 0.1;
    
    // Projectiles currently in the simulation
    list<Projectile*> m_proj;
    
    // Collisions generated in the last time unit (i.e the last call to advance()).
    list <Collision> m_collisions;
    
    // World boundaries
    GLfloat m_xMax;
    GLfloat m_yMax;
};













#endif // __PROJ_MODEL_H__